"""
Subpackage used for analyzing data collected by catalight automation tools.

Plotting, user dialogs, and scripts contained within. Note: any module prefixed
'run\\_' can be executed directly to utilize the a simple gui dialog or used by
calling module.main() function.
"""
from . import tools  # noqa
from . import plotting  # noqa
from . import user_inputs  # noqa
